<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

namespace tool_userupsert\external;

use tool_userupsert\event\upsert_failed;
use tool_userupsert\config;
use core_external\external_api;
use core_external\external_function_parameters;
use core_external\external_single_structure;
use core_external\external_multiple_structure;
use core_external\external_value;
use tool_userupsert\event\upsert_succeeded;
use tool_userupsert\user_manager;

/**
 * Web Service functions.
 *
 * @package     tool_userupsert
 * @copyright   2021 Catalyst IT
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class upsert extends external_api {

    /**
     * Define the parameters schema for the upsert_user function.
     *
     * @return external_function_parameters
     */
    public static function upsert_users_parameters(): external_function_parameters {
        $config = new config();
        $userfields = [];
        $extradescription = ' (' . get_string('requiredtocreateuser', 'tool_userupsert') . ')';

        foreach ($config->get_web_service_fields() as $field => $description) {
            $required = false;
            $mappedfields = array_keys($config->get_data_mapping(), $field);

            foreach ($mappedfields as $fieldname) {

                // Check if one of the  mapped fields is mandatory field.
                if (in_array($fieldname, $config->get_mandatory_fields())) {
                    $description .= $extradescription;
                }

                // Check if mapped field is being used for matching users.
                if ($fieldname == $config->get_user_match_field()) {
                    $required = true;
                }
            }

            if ($required) {
                $userfields[$field] = new external_value(PARAM_RAW, $description, VALUE_REQUIRED);
            } else {
                $userfields[$field] = new external_value(PARAM_RAW, $description, VALUE_OPTIONAL);
            }
        }

        return new external_function_parameters([
            'users' => new external_multiple_structure(new external_single_structure($userfields))
        ]);
    }

    /**
     * Upserts users.
     *
     * @param array $users A list of user fields.
     * @return array
     */
    public static function upsert_users(array $users): array {
        $params = self::validate_parameters(self::upsert_users_parameters(), ['users' => $users]);
        require_capability('tool/userupsert:upsert', \context_system::instance());

        $warnings = [];

        $config = new config();
        $usermanager = new user_manager($config);

        foreach ($params['users'] as $user) {

            if (!empty($user[$config->get_data_mapping()[$config->get_user_match_field()]])) {
                $itemid = $user[$config->get_data_mapping()[$config->get_user_match_field()]];
            } else {
                $itemid = 'not set';
            }

            try {
                $usermanager->upsert_user($user);
                upsert_succeeded::create([
                    'other' => [
                        'itemid' => $itemid,
                    ]
                ])->trigger();
            } catch (\Exception $exception) {
                $warning = [
                    'itemid' => $itemid,
                    'error' => $exception->getMessage(),
                ];

                $warnings[] = $warning;
                upsert_failed::create(['other' => $warning])->trigger();
            }
        }

        return $warnings;
    }

    /**
     * Return values for upsert_user external function.
     *
     * @return external_multiple_structure
     */
    public static function upsert_users_returns(): external_multiple_structure {
        return new external_multiple_structure(
            new external_single_structure(
                [
                    'itemid' => new external_value(PARAM_RAW, 'User matching identificator', VALUE_OPTIONAL),
                    'error' => new external_value(PARAM_RAW, 'Message to explain the error', VALUE_OPTIONAL)
                ]
            )
        );
    }

}
